;; Generated package description from package-x.el  -*- no-byte-compile: t -*-
(define-package "package-x" "1.0.0.20250516.155254" "Package extras" 'nil :commit "00f5b5f5a08a3fb2797f614eb084cfa3de91fa8a" :url "https://elpa.gnu.org/packages/package-x.html" :authors '(("Tom Tromey" . "tromey@redhat.com")) :maintainer '("Philip Kaludercic" . "philipk@posteo.net") :keywords '("tools"))
